/* Interactive Firewall for Mandriva
 * Samir Bellabes <sbellabes@mandriva.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * 2007-10-10 Thomas Backlund <tmb@mandriva.org>: build fixes for 2.6.22.9
 * 2007-11-11 Herton Krzesinski <herton@mandriva.com>: build fixes for 2.6.24-rc
 * 2007-12-03 Luiz Capitulino <lcapitulino@mandriva.com.br>: v1.1
 * 		- Better multicast group usage
 * 		- Coding style fixes
 * 		- Do not return -EINVAL by default in ipt_ifwlog_init()
 * 		- Minor refinements
 */

#include <linux/module.h>
#include <asm/types.h>
#include <linux/jiffies.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/icmp.h>
#include <net/udp.h>
#include <net/tcp.h>
#include <net/sock.h>
#include <linux/netlink.h>
#include <linux/string.h>

#include <linux/netfilter.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_IFWLOG.h>


#if 0
#define DEBUGP PRINTR
#else
#define DEBUGP(format, args...)
#endif

#define PRINTR(format, args...) do { if(net_ratelimit()) printk(format, ##args); } while(0)

static struct sock *nl;

/* send struct to userspace */
static void send_packet(const struct nl_msg *msg)
{
	struct sk_buff *skb = NULL;
	struct nlmsghdr *nlh;
	unsigned int size;

	size = NLMSG_SPACE(sizeof(*msg));
	skb = alloc_skb(size, GFP_ATOMIC);
	if (!skb) {
		PRINTR(KERN_WARNING "IFWLOG: OOM can't allocate skb\n");
		return;
	}

	nlh = nlmsg_put(skb, 0, 0, 0, size - sizeof(*nlh), 0);

	memcpy(nlmsg_data(nlh), (const void *) msg, sizeof(*msg));

	NETLINK_CB(skb).portid = 0;  /* from kernel */
	NETLINK_CB(skb).dst_group = IFWLOGNLGRP_DEF;

	if (nl) {
		DEBUGP(KERN_WARNING
		       "IFWLOG: nlmsg_len=%ld\nnlmsg_type=%d nlmsg_flags=%d\nnlmsg_seq=%ld nlmsg_pid = %ld\n",
		       (long)nlh->nlmsg_len,  nlh->nlmsg_type, nlh->nlmsg_flags,
		       (long)nlh->nlmsg_seq, (long)nlh->nlmsg_pid);
		DEBUGP(KERN_WARNING "prefix : %s\n", msg->prefix);

		netlink_broadcast(nl, skb, 0, IFWLOGNLGRP_DEF, GFP_ATOMIC);
		return;
	}

nlmsg_failure:
	kfree_skb(skb);
	PRINTR(KERN_WARNING "IFWLOG: Error sending netlink packet\n");
}

/* fill struct for userspace */
static void ipt_IFWLOG_packet(const struct sk_buff *skb,
			      const struct net_device *in,
			      const struct net_device *out,
			      const struct ipt_IFWLOG_info *info)
{
	struct iphdr iph;
	struct tcphdr tcph;
	struct udphdr udph;
	struct nl_msg msg;
	struct iphdr _iph, *ih;
	struct timeval tv;

	memset(&msg, 0, sizeof(struct nl_msg));

	ih = skb_header_pointer(skb, 0, sizeof(_iph), &_iph);
	if (ih == NULL) {
		PRINTR(KERN_WARNING "IFWLOG: skb truncated");
		return;
	}

	/* save interface name */
	if (in)
		strcpy(msg.indev_name, in->name);
	if (out)
		strcpy(msg.outdev_name, out->name);

	/* save log-prefix */
	strcpy(msg.prefix, info->prefix);

	/* save ip header */
	skb_copy_bits(skb, 0, &iph, sizeof(iph));
	memcpy(&msg.ip, &iph, sizeof(struct iphdr));

	/* save transport header */
	switch (iph.protocol){
	case IPPROTO_TCP:
		skb_copy_bits(skb, iph.ihl*4 , &tcph, sizeof(tcph));
		memcpy(&msg.h.th, &tcph, sizeof(struct tcphdr));
		break;
	case IPPROTO_UDP:
		skb_copy_bits(skb, iph.ihl*4 , &udph, sizeof(udph));
		memcpy(&msg.h.uh, &udph, sizeof(struct udphdr));
		break;
	default:
		break;
	}

	/* save timetamp */
	do_gettimeofday((struct timeval *)&tv);
        msg.timestamp_sec = tv.tv_sec;

	send_packet(&msg);
}

static unsigned int ipt_IFWLOG_target(struct sk_buff *skb,
				      const struct xt_action_param *target_param)
{
	const struct ipt_IFWLOG_info *info = target_param->targinfo;

	ipt_IFWLOG_packet(skb, target_param->in, target_param->out, info);

	return XT_CONTINUE;
}

static int ipt_IFWLOG_checkentry(const struct xt_tgchk_param *tgchk_param)
{
	const struct ipt_IFWLOG_info *info = tgchk_param->targinfo;

	if (info->prefix[sizeof(info->prefix)-1] != '\0') {
		DEBUGP("IFWLOG: prefix term %i\n",
		       info->prefix[sizeof(info->prefix)-1]);
		return -EINVAL;
	}

	return 0;
}

static struct xt_target ipt_IFWLOG = {
	.name		= "IFWLOG",
	.family		= AF_INET,
	.target		= ipt_IFWLOG_target,
	.targetsize	= sizeof(struct ipt_IFWLOG_info),
	.checkentry	= ipt_IFWLOG_checkentry,
	.me		= THIS_MODULE,
};

static int __init ipt_ifwlog_init(void)
{
	int err;
	struct netlink_kernel_cfg cfg = {
		.groups = IFWLOGNLGRP_MAX,
	};

	nl = netlink_kernel_create(&init_net, NETLINK_IFWLOG, &cfg);
	if (!nl) {
		PRINTR(KERN_WARNING "IFWLOG: cannot create netlink socket\n");
		return -ENOMEM;
	}

	err = xt_register_target(&ipt_IFWLOG);
	if (err) {
		if (nl && nl->sk_socket)
			sock_release(nl->sk_socket);
		return err;
	}

	PRINTR(KERN_INFO "IFWLOG: register target\n");
	return 0;
}

static void __exit ipt_ifwlog_fini(void)
{
	if (nl && nl->sk_socket)
		sock_release(nl->sk_socket);
	PRINTR(KERN_INFO "IFWLOG: unregister target\n");
	xt_unregister_target(&ipt_IFWLOG);
}

module_init(ipt_ifwlog_init);
module_exit(ipt_ifwlog_fini);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Samir Bellabes <sbellabes@mandriva.com>");
MODULE_AUTHOR("Luiz Capitulino <lcapitulino@mandriva.com.br>");
MODULE_DESCRIPTION("Interactive firewall logging and module");
MODULE_VERSION("v1.1");
